#!/usr/bin/env python

"""
Plot ancestry PCA analysis results
"""

import logging
from collections import Counter
from typing import List, Iterable, Optional

import click
import pandas as pd
import numpy as np
import hail as hl
from bokeh.resources import CDN
from bokeh.embed import file_html
from bokeh.transform import factor_cmap, factor_mark
from bokeh.plotting import ColumnDataSource, figure
from bokeh.palettes import turbo  # pylint: disable=no-name-in-module
from bokeh.models import CategoricalColorMapper, HoverTool

from joint_calling import utils, resources
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--eigenvalues',
    'eigenvalues_path',
    required=True,
    callback=utils.get_validation_callback(ext='txt', must_exist=True),
)
@click.option(
    '--scores-ht',
    'scores_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--loadings-ht',
    'loadings_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--provided-pop-ht',
    'provided_pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--inferred-pop-ht',
    'inferred_pop_ht_path',
    required=True,
    callback=utils.get_validation_callback(ext='ht', must_exist=True),
)
@click.option(
    '--meta-tsv',
    'meta_tsv_path',
    required=True,
    callback=utils.get_validation_callback(ext='tsv', must_exist=True),
)
@click.option('--out-path-pattern', 'out_path_pattern')
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    eigenvalues_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    provided_pop_ht_path: str,
    inferred_pop_ht_path: str,
    meta_tsv_path: str,
    out_path_pattern: Optional[str],
    hail_billing: str,  # pylint: disable=unused-argument
):
    local_tmp_dir = utils.init_hail(__file__)

    meta_ht = utils.parse_input_metadata(meta_tsv_path, local_tmp_dir)

    produce_plots(
        eigenvalues_path=eigenvalues_path,
        scores_ht_path=scores_ht_path,
        loadings_ht_path=loadings_ht_path,
        provided_pop_ht_path=provided_pop_ht_path,
        inferred_pop_ht_path=inferred_pop_ht_path,
        meta_ht=meta_ht,
        out_path_pattern=out_path_pattern,
    )


def key_by_external_id(ht, sample_map_ht):
    """
    Assuming ht.s is a CPG id, replaces it with external ID, using map in sample_map_ht
    """
    ht = ht.annotate(internal_id=ht.s).key_by('internal_id')
    ht = (
        ht.annotate(
            s=hl.if_else(
                hl.is_defined(sample_map_ht[ht.internal_id]),
                sample_map_ht[ht.internal_id].external_id,
                ht.internal_id,
            )
        )
        .key_by('s')
        .drop('internal_id')
    )
    return ht


def produce_plots(
    eigenvalues_path: str,
    scores_ht_path: str,
    loadings_ht_path: str,
    provided_pop_ht_path: str,
    inferred_pop_ht_path: str,
    meta_ht: hl.Table,
    out_path_pattern: Optional[str] = None,
    number_of_pcs: Optional[int] = None,
):
    """
    Generate plots in HTML format, write for each PC (of n_pcs) and
    scope ("study", "continental_pop", "subpop", plus for loadings) into
    file paths defined by `out_path_pattern`.
    """
    scores_ht = hl.read_table(scores_ht_path)
    provided_pop_ht = hl.read_table(provided_pop_ht_path)
    inferred_pop_ht = hl.read_table(inferred_pop_ht_path)

    sample_map_ht = meta_ht.select('external_id')
    scores_ht = key_by_external_id(scores_ht, sample_map_ht).cache()
    provided_pop_ht = key_by_external_id(provided_pop_ht, sample_map_ht).cache()
    inferred_pop_ht = key_by_external_id(inferred_pop_ht, sample_map_ht).cache()

    scores_ht = scores_ht.annotate(
        continental_pop=hl.case()
        .when(
            provided_pop_ht[scores_ht.s].continental_pop != '',
            provided_pop_ht[scores_ht.s].project
            + ' ['
            + provided_pop_ht[scores_ht.s].continental_pop
            + ']',
        )
        .default(
            provided_pop_ht[scores_ht.s].project
            + ' [inferred '
            + inferred_pop_ht[scores_ht.s].pop
            + ']'
        ),
        subpop=hl.case()
        .when(
            provided_pop_ht[scores_ht.s].subpop != '',
            provided_pop_ht[scores_ht.s].project
            + ' ['
            + provided_pop_ht[scores_ht.s].subpop
            + ']',
        )
        .default(
            provided_pop_ht[scores_ht.s].project
            + ' [inferred '
            + inferred_pop_ht[scores_ht.s].pop
            + ']'
        ),
        study=provided_pop_ht[scores_ht.s].project,
    ).cache()

    eigenvalues_ht = hl.import_table(
        eigenvalues_path, no_header=True, types={'f0': hl.tfloat}
    ).f0.collect()
    eigenvalues_ht = pd.to_numeric(eigenvalues_ht)
    variance = np.divide(eigenvalues_ht[1:], float(eigenvalues_ht.sum())) * 100
    variance = variance.round(2)
    number_of_pcs = number_of_pcs or len(eigenvalues_ht) - 1

    plots = []

    sample_names = scores_ht.s.collect()
    studies = scores_ht.study.collect()
    unique_studies = remove_duplicates(studies)
    bg_studies = [s for s in unique_studies if s == 'gnomad']
    fg_studies = [s for s in unique_studies if s != 'gnomad']

    plots.extend(
        _plot_study(
            number_of_pcs,
            variance,
            scores_ht,
            studies,
            sample_names,
            bg_studies,
            fg_studies,
            out_path_pattern=out_path_pattern,
        )
    )

    plots.extend(
        _plot_continental_pop(
            number_of_pcs,
            variance,
            scores_ht,
            studies,
            sample_names,
            bg_studies,
            fg_studies,
            out_path_pattern=out_path_pattern,
        )
    )

    plots.extend(
        _plot_subcontinental_pop(
            number_of_pcs,
            variance,
            scores_ht,
            studies,
            sample_names,
            bg_studies,
            fg_studies,
            out_path_pattern=out_path_pattern,
        )
    )

    plots.extend(
        _plot_loadings(
            number_of_pcs, loadings_ht_path, out_path_pattern=out_path_pattern
        )
    )

    return plots


def _plot_study(
    number_of_pcs,
    variance,
    scores,
    studies,
    sample_names,
    bg_studies,
    fg_studies,
    out_path_pattern=None,
):
    tooltips = [('labels', '@label'), ('samples', '@samples')]
    cntr: Counter = Counter(studies)
    labels = [f'{x} ({cntr[x]})' for x in studies]
    unique_labels = list(Counter(labels).keys())

    plots = []
    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Study',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
            width=1000,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
                study=studies,
            )
        )
        plot.scatter(
            'x',
            'y',
            alpha=0.5,
            marker=factor_mark(
                'study',
                ['cross'] * len(bg_studies) + ['circle'] * len(fg_studies),
                bg_studies + fg_studies,
            ),
            source=source,
            size=4,
            color=factor_cmap('label', ['#1b9e77', '#d95f02'], unique_labels),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, 'my plot')
            plot_filename_html = out_path_pattern.format(
                scope='study', pci=pc2, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def _plot_continental_pop(
    number_of_pcs,
    variance,
    scores,
    studies,
    sample_names,
    bg_studies,
    fg_studies,
    out_path_pattern=None,
):
    tooltips = [('labels', '@label'), ('samples', '@samples')]
    labels = scores.continental_pop.collect()
    cntr = Counter(labels)
    labels = [f'{x} ({cntr[x]})' for x in labels]
    unique_labels = list(Counter(labels).keys())

    plots = []
    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Continental Population',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
            width=1000,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
                study=studies,
            )
        )
        plot.scatter(
            'x',
            'y',
            alpha=0.6,
            marker=factor_mark(
                'study',
                ['cross'] * len(bg_studies) + ['circle'] * len(fg_studies),
                bg_studies + fg_studies,
            ),
            source=source,
            size=4,
            color=factor_cmap('label', turbo(len(unique_labels)), unique_labels),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, 'my plot')
            plot_filename_html = out_path_pattern.format(
                scope='continental_pop', pci=pc2, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def _plot_subcontinental_pop(
    number_of_pcs,
    variance,
    scores,
    studies,
    sample_names,
    bg_studies,
    fg_studies,
    out_path_pattern=None,
):
    tooltips = [('labels', '@label'), ('samples', '@samples')]
    labels = scores.subpop.collect()
    cntr = Counter(labels)
    labels = [f'{x} ({cntr[x]})' for x in labels]
    unique_labels = list(Counter(labels).keys())

    plots = []
    for i in range(number_of_pcs - 1):
        pc1 = i
        pc2 = i + 1
        plot = figure(
            title='Subpopulation',
            x_axis_label=f'PC{pc1 + 1} ({variance[pc1]})%)',
            y_axis_label=f'PC{pc2 + 1} ({variance[pc2]}%)',
            tooltips=tooltips,
            width=1000,
        )
        source = ColumnDataSource(
            dict(
                x=scores.scores[pc1].collect(),
                y=scores.scores[pc2].collect(),
                label=labels,
                samples=sample_names,
                study=studies,
            )
        )
        plot.scatter(
            'x',
            'y',
            alpha=0.6,
            marker=factor_mark(
                'study',
                ['cross'] * len(bg_studies) + ['circle'] * len(fg_studies),
                bg_studies + fg_studies,
            ),
            source=source,
            size=4,
            color=factor_cmap('label', turbo(len(unique_labels)), unique_labels),
            legend_group='label',
        )
        plot.add_layout(plot.legend[0], 'left')
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, 'my plot')
            plot_filename_html = out_path_pattern.format(
                scope='subpop', pci=pc2, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def _plot_loadings(number_of_pcs, loadings_ht_path, out_path_pattern=None):
    loadings_ht = hl.read_table(loadings_ht_path)
    plots = []
    gtf_ht = hl.experimental.import_gtf(
        resources.GENCODE_GTF,
        reference_genome='GRCh38',
        skip_invalid_contigs=True,
        min_partitions=12,
        force_bgz=True,
    )
    for i in range(number_of_pcs - 1):
        pc = i + 1
        plot = manhattan_loadings(
            iteration=i,
            gtf=gtf_ht,
            loadings=loadings_ht,
            title='Loadings of PC ' + str(pc),
            collect_all=True,
        )
        plots.append(plot)
        if out_path_pattern:
            html = file_html(plot, CDN, 'my plot')
            plot_filename_html = out_path_pattern.format(
                scope='loadings', pci=pc, ext='html'
            )
            with hl.hadoop_open(plot_filename_html, 'w') as f:
                f.write(html)
    return plots


def manhattan_loadings(
    iteration,
    gtf,
    loadings,
    title=None,
    size=4,
    hover_fields=None,
    collect_all=False,
    n_divisions=500,
):
    """modify hail manhattan plot"""
    palette = [
        '#1f77b4',
        '#ff7f0e',
        '#2ca02c',
        '#d62728',
        '#9467bd',
        '#8c564b',
        '#e377c2',
        '#7f7f7f',
        '#bcbd22',
        '#17becf',
    ]

    # add gene names, p-values, and locus info
    loadings = loadings.annotate(gene_names=gtf[loadings.locus].gene_name)
    pvals = hl.abs(loadings.loadings[iteration])
    locus = loadings.locus

    if hover_fields is None:
        hover_fields = {}

    hover_fields['locus'] = hl.str(locus)
    hover_fields['gene'] = hl.str(loadings.gene_names)

    source_pd = (
        hl.plot.plots._collect_scatter_plot_data(  # pylint: disable=protected-access
            ('_global_locus', locus.global_position()),
            ('_pval', pvals),
            fields=hover_fields,
            n_divisions=None if collect_all else n_divisions,
        )
    )
    source_pd['p_value'] = source_pd['_pval']
    source_pd['_contig'] = [locus.split(':')[0] for locus in source_pd['locus']]

    observed_contigs = set(source_pd['_contig'])
    ref = locus.dtype.reference_genome
    observed_contigs = [
        contig for contig in ref.contigs.copy() if contig in observed_contigs
    ]

    contig_ticks = [
        ref._contig_global_position(contig)  # pylint: disable=protected-access
        + ref.contig_length(contig) // 2
        for contig in observed_contigs
    ]
    color_mapper = CategoricalColorMapper(
        factors=ref.contigs, palette=palette[:2] * int((len(ref.contigs) + 1) / 2)
    )

    p = figure(
        title=title, x_axis_label='Chromosome', y_axis_label='Loadings', width=1000
    )
    (
        p,
        _,
        legend,
        _,
        _,
        _,
    ) = hl.plot.plots._get_scatter_plot_elements(  # pylint: disable=protected-access
        p,
        source_pd,
        x_col='_global_locus',
        y_col='_pval',
        label_cols=['_contig'],
        colors={'_contig': color_mapper},
        size=size,
    )
    legend.visible = False
    p.xaxis.ticker = contig_ticks
    p.xaxis.major_label_overrides = dict(zip(contig_ticks, observed_contigs))
    p.select_one(HoverTool).tooltips = [
        t for t in p.select_one(HoverTool).tooltips if not t[0].startswith('_')
    ]

    return p


def remove_duplicates(x: Iterable) -> List:
    """
    Removes duplicates from a list, keeps order
    """
    return list(dict.fromkeys(x))


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Dataproc cluster utils
"""
from typing import Optional, List
from analysis_runner import dataproc
from hailtop.batch.job import Job
from hailtop.batch import Batch
from joint_calling import utils


DataprocCluster = dataproc.DataprocCluster


def get_cluster(
    b: Batch,
    name,
    num_workers: int,
    long: bool = False,
    is_test: bool = False,
    phantomjs: bool = False,
    preemptible: bool = True,
    depends_on: Optional[List[Job]] = None,
) -> dataproc.DataprocCluster:
    """
    Get or create a Dataproc cluster by name
    """
    max_age = '1h' if is_test else '8h'
    if long:
        max_age = '3h' if is_test else '24h'

    return dataproc.setup_dataproc(
        b,
        max_age=max_age,
        packages=utils.DATAPROC_PACKAGES,
        num_secondary_workers=num_workers if preemptible else 0,
        num_workers=2 if preemptible else num_workers,
        cluster_name=name,
        depends_on=depends_on,
        init=['gs://cpg-reference/hail_dataproc/install_phantomjs.sh']
        if phantomjs
        else [],
    )



#!/usr/bin/env python3

"""
This script parses "somalier relate" (https://github.com/brentp/somalier) outputs,
and returns a non-zero code if either sex or pedigree mismatches the data in a 
provided PED file.
"""

import contextlib
import logging
import subprocess
import sys
import tempfile
from os.path import join, basename
from typing import Dict, List

import pandas as pd
import click
from peddy import Ped


logger = logging.getLogger('check-pedigree')
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.option(
    '--somalier-samples',
    'somalier_samples_fpath',
    required=True,
    help='Path to somalier {prefix}.samples.tsv output file',
)
@click.option(
    '--somalier-pairs',
    'somalier_pairs_fpath',
    required=True,
    help='Path to somalier {prefix}.pairs.tsv output file',
)
@click.option(
    '--somalier-html',
    'somalier_html_fpath',
    help='Path to somalier {prefix}.html output file',
)
def main(
    somalier_samples_fpath: str,
    somalier_pairs_fpath: str,
    somalier_html_fpath: str,
):  # pylint: disable=missing-function-docstring
    if somalier_samples_fpath.startswith('gs://'):
        local_tmp_dir = tempfile.mkdtemp()
        cmd = f'gsutil cp {somalier_samples_fpath} {somalier_pairs_fpath}'
        subprocess.run(cmd, check=False, shell=True)
        local_somalier_pairs_fpath = join(local_tmp_dir, basename(somalier_pairs_fpath))
        local_somalier_samples_fpath = join(
            local_tmp_dir, basename(somalier_samples_fpath)
        )
    else:
        local_somalier_pairs_fpath = somalier_pairs_fpath
        local_somalier_samples_fpath = somalier_samples_fpath

    logger.info('* Checking sex *')
    df = pd.read_csv(local_somalier_samples_fpath, delimiter='\t')
    missing_sex = df['original_pedigree_sex'] == 'unknown'
    mismatching_female = (df['sex'] == 2) & (df['original_pedigree_sex'] == 'male')
    mismatching_male = (df['sex'] == 1) & (df['original_pedigree_sex'] == 'female')
    mismatching_sex = mismatching_female | mismatching_male

    if (mismatching_sex | missing_sex).any():
        logger.info(f'Found PED samples with mismatching or missing sex:')
        for _, row in df[mismatching_sex | missing_sex].iterrows():
            inferred_sex = {1: 'male', 2: 'female'}.get(row.sex, 'unknown')
            logger.info(
                f'\t{row.sample_id} (provided: {row.original_pedigree_sex}, inferred: {inferred_sex})'
            )

    if not mismatching_sex.any():
        if missing_sex.any():
            logger.info(
                f'Inferred sex and pedigree matches or can be inferred for all samples.'
            )
        else:
            logger.info(f'Inferred sex and pedigree matches for all samples.')
    logger.info('-' * 10)

    logger.info('* Checking relatedness *')
    ped = Ped(local_somalier_samples_fpath)
    sample_by_id = {s.sample_id: s for s in ped.samples()}
    mismatching_pairs = []
    pairs_df = pd.read_csv(local_somalier_pairs_fpath, delimiter='\t')
    for idx, row in pairs_df.iterrows():
        s1 = row['#sample_a']
        s2 = row['sample_b']
        inferred_rel = infer_relationship(row['relatedness'], row['ibs0'], row['ibs2'])
        # Supressing all logging output from peddy as it would clutter the logs
        with contextlib.redirect_stderr(None):
            with contextlib.redirect_stdout(None):
                provided_rel = ped.relation(sample_by_id[s1], sample_by_id[s2])

        provided_to_inferred: Dict[str, List[str]] = {
            'unrelated': ['unrelated'],
            'related at unknown level': ['unrelated'],  # e.g. mom-dad
            'mom-dad': ['unrelated'],
            'parent-child': ['parent-child'],
            'grandchild': ['below_first_degree'],
            'niece/nephew': ['below_first_degree'],
            'great-grandchild': ['below_first_degree'],
            'cousins': ['below_first_degree'],
            'full siblings': ['siblings'],
            'siblings': ['siblings'],
            'unknown': [],
        }
        if inferred_rel not in provided_to_inferred[provided_rel]:
            mismatching_pairs.append(
                f'"{s1}" and "{s2}", '
                f'provided relationship: "{provided_rel}", inferred: "{inferred_rel}"'
            )
        pairs_df.loc[idx, 'provided_rel'] = provided_rel
        pairs_df.loc[idx, 'inferred_rel'] = inferred_rel

    if mismatching_pairs:
        logger.info(f'Found sample pairs with mismatched relatedness:')
        for pair in mismatching_pairs:
            logger.info(f'\t{pair}')
    else:
        logger.info(f'Inferred pedigree matches for all samples.')

    logger.info('-' * 10)
    print_info(
        df,
        pairs_df,
        somalier_samples_fpath,
        somalier_pairs_fpath,
        somalier_html_fpath,
    )
    if mismatching_sex.any() or mismatching_pairs:
        sys.exit(1)


def infer_relationship(coeff: float, ibs0: float, ibs2: float) -> str:
    """
    Inferres relashionship labels based on the kin coefficient
    and ibs0 and ibs2 values.
    """
    result = 'ambiguous'
    if coeff < 0.2:
        result = 'unrelated'
    elif 0.2 <= coeff < 0.38:
        result = 'below_first_degree'
    elif 0.38 <= coeff <= 0.62:
        if ibs0 / ibs2 < 0.005:
            result = 'parent-child'
        elif 0.015 < ibs0 / ibs2 < 0.052:
            result = 'siblings'
        else:
            result = 'first_degree'
    elif coeff > 0.8:
        result = 'duplicate_or_twins'
    return result


def print_info(
    samples_df,
    pairs_df,
    somalier_samples_fpath,
    somalier_pairs_fpath,
    somalier_html_fpath,
):
    """
    Print useful information to manually review pedigree check results
    """
    if len(samples_df) < 400:
        samples_str = samples_df.to_string()
        logger.info('')
        logger.info(
            f'Somalier results, samples (based on {somalier_samples_fpath}):\n{samples_str}\n'
        )
    if len(pairs_df) < 400:
        pairs_str = pairs_df[
            [
                '#sample_a',
                'sample_b',
                'relatedness',
                'ibs0',
                'ibs2',
                'n',
                'expected_relatedness',
            ]
        ].to_string()
        logger.info(
            f'Somalier results, sample pairs (based on {somalier_pairs_fpath}):\n{pairs_str}\n'
        )
    if somalier_html_fpath:
        logger.info(f'Somalier HTML report: {somalier_html_fpath}\n')


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Utility module to encapsulate pointers to the reference files used in the pipeline
"""

import functools
import logging
import operator
from os.path import join
from typing import List, Optional, Union

import hail as hl


logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


REF_BUCKET = 'gs://cpg-reference'
NOALT_REGIONS = join(REF_BUCKET, 'noalt.bed')
SOMALIER_SITES = join(REF_BUCKET, 'somalier/v0/sites.hg38.vcf.gz')

BROAD_REF_BUCKET = f'{REF_BUCKET}/hg38/v1'
REF_FASTA = join(BROAD_REF_BUCKET, 'Homo_sapiens_assembly38.fasta')
UNPADDED_INTERVALS = join(BROAD_REF_BUCKET, 'hg38.even.handcurated.20k.intervals')

GNOMAD_REF_BUCKET = f'{REF_BUCKET}/gnomad/v0'
TEL_AND_CENT_HT = join(
    GNOMAD_REF_BUCKET,
    'telomeres_and_centromeres/hg38.telomeresAndMergedCentromeres.ht',
)
LCR_INTERVALS_HT = join(GNOMAD_REF_BUCKET, 'lcr_intervals/LCRFromHengHg38.ht')
SEG_DUP_INTERVALS_HT = join(GNOMAD_REF_BUCKET, 'seg_dup_intervals/GRCh38_segdups.ht')
CLINVAR_HT = join(GNOMAD_REF_BUCKET, 'clinvar/clinvar_20190923.ht')
HAPMAP_HT = join(GNOMAD_REF_BUCKET, 'hapmap/hapmap_3.3.hg38.ht')
KGP_OMNI_HT = join(GNOMAD_REF_BUCKET, 'kgp/1000G_omni2.5.hg38.ht')
KGP_HC_HT = join(GNOMAD_REF_BUCKET, 'kgp/1000G_phase1.snps.high_confidence.hg38.ht')
MILLS_HT = join(
    GNOMAD_REF_BUCKET, 'mills/Mills_and_1000G_gold_standard.indels.hg38.ht/'
)

GENCODE_GTF = join(REF_BUCKET, 'gencode/gencode.v29.annotation.gtf.bgz')

GNOMAD_HT = (
    'gs://gcp-public-data--gnomad/release/3.1/ht/genomes/gnomad.genomes.v3.1.sites.ht'
)

ANCESTRY_BUCKET = f'{REF_BUCKET}/hg38/ancestry/v3-90k'
ANCESTRY_HGDP_SUBSET_MTS = {
    'all': f'{ANCESTRY_BUCKET}/gnomad_subset.mt',
    'nfe': f'{ANCESTRY_BUCKET}/gnomad_subset_nfe.mt',
    'test': f'{ANCESTRY_BUCKET}/gnomad_subset_test.mt',
    'test_nfe': f'{ANCESTRY_BUCKET}/gnomad_subset_test_nfe.mt',
}
ANCESTRY_SITES = f'{ANCESTRY_BUCKET}/pca_sites.ht'


def filter_low_conf_regions(
    mt: Union[hl.MatrixTable, hl.Table],
    filter_lcr: bool = True,
    filter_segdup: bool = True,
    filter_telomeres_and_centromeres: bool = False,
    high_conf_regions: Optional[List[str]] = None,
) -> Union[hl.MatrixTable, hl.Table]:
    """
    Filter low-confidence regions.

    :param mt: MatrixTable or Table to filter
    :param filter_lcr: Whether to filter LCR regions
    :param filter_segdup: Whether to filter Segdup regions
    :param filter_telomeres_and_centromeres: Whether to filter telomeres and centromeres
    :param high_conf_regions: Paths to set of high confidence regions to restrict to (union of regions)
    :return: MatrixTable or Table with low confidence regions removed
    """
    criteria = []
    if filter_lcr:
        lcr = hl.read_table(LCR_INTERVALS_HT)
        criteria.append(hl.is_missing(lcr[mt.locus]))

    if filter_segdup:
        segdup = hl.read_table(SEG_DUP_INTERVALS_HT)
        criteria.append(hl.is_missing(segdup[mt.locus]))

    if filter_telomeres_and_centromeres:
        telomeres_and_centromeres = hl.read_table(TEL_AND_CENT_HT)
        criteria.append(hl.is_missing(telomeres_and_centromeres[mt.locus]))

    if high_conf_regions is not None:
        for region in high_conf_regions:
            region = hl.import_locus_intervals(region)
            criteria.append(hl.is_defined(region[mt.locus]))

    if criteria:
        filter_criteria = functools.reduce(operator.iand, criteria)
        if isinstance(mt, hl.MatrixTable):
            mt = mt.filter_rows(filter_criteria)
        else:
            mt = mt.filter(filter_criteria)

    return mt


def get_truth_ht() -> hl.Table:
    """
    Return a table with annotations from the latest version of the corresponding truth data.

    The following annotations are included:
        - hapmap
        - kgp_omni (1000 Genomes intersection Onni 2.5M array)
        - kgp_phase_1_hc (high confidence sites in 1000 genonmes)
        - mills (Mills & Devine indels)

    :return: A table with the latest version of popular truth data annotations
    """
    return (
        hl.read_table(HAPMAP_HT)
        .select(hapmap=True)
        .join(hl.read_table(KGP_OMNI_HT).select(omni=True), how='outer')
        .join(hl.read_table(KGP_HC_HT).select(kgp_phase1_hc=True), how='outer')
        .join(hl.read_table(MILLS_HT).select(mills=True), how='outer')
        .repartition(200, shuffle=False)
        .persist()
    )



#!/usr/bin/env python

"""
Convert somalier pairwise relatedness file into a Hail table similar to the 
one generated by hl.pc_relate, expected by the downstream analysis
"""

import logging
import pandas as pd
import click
import hail as hl

from joint_calling import utils
from joint_calling import _version


logger = logging.getLogger(__file__)
logging.basicConfig(format='%(levelname)s (%(name)s %(lineno)s): %(message)s')
logger.setLevel(logging.INFO)


@click.command()
@click.version_option(_version.__version__)
@click.option(
    '--somalier-pairs-tsv',
    'somalier_pairs_tsv_path',
    required=True,
)
@click.option(
    '--out-relatedness-ht',
    'out_relatedness_ht_path',
)
@click.option(
    '--overwrite/--reuse',
    'overwrite',
    is_flag=True,
    help='if an intermediate or a final file exists, skip running the code '
    'that generates it.',
)
@click.option(
    '--hail-billing',
    'hail_billing',
    required=True,
    help='Hail billing account ID.',
)
def main(  # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    somalier_pairs_tsv_path: str,
    out_relatedness_ht_path: str,
    overwrite: bool,
    hail_billing: str,  # pylint: disable=unused-argument
):
    utils.init_hail(__file__)

    if not utils.can_reuse(out_relatedness_ht_path, overwrite):
        pairs_df = pd.read_csv(somalier_pairs_tsv_path, delimiter='\t')
        pairs_df = pairs_df.rename(
            columns={'#sample_a': 'i', 'sample_b': 'j', 'relatedness': 'kin'}
        )
        ht = hl.Table.from_pandas(pairs_df).key_by('i', 'j')
        ht.write(out_relatedness_ht_path, overwrite=True)


if __name__ == '__main__':
    main()  # pylint: disable=E1120



"""
Prepare input tsv file for nagim to run with the workflow as follows:
batch_workflow.py --input-tsv gs://cpg-nagim-test/joint_calling/samples.tsv
"""

import os
import pandas as pd
from joint_calling import utils

def run_cmd(cmd):
    print(cmd)
    os.system(cmd)

run_cmd('gsutil ls \'gs://cpg-nagim-test-upload/gvcf/*.vcf.gz\' > gvcflist.txt')

default_entry = {
    's': None,
    'external_id': None,
    'project': None,
    'continental_pop': '-',
    'subpop': '-',
    'gvcf': '-',
    'topostproc_gvcf': '-',
    'cram': '-',
    'crai': '-',
    'realign_cram': '-',
    'realign_crai': '-',
    'batch': '-',
    'operation': 'add',
    'flowcell_lane': '-',
    'library_id': '-',
    'platform': '-',
    'centre': '-',
    'r_contamination': None,
    'r_chimera': None,
    'r_duplication': None,
    'median_insert_size': None,
    'fam_id': '-',
    'mat_id': '-',
    'pat_id': '-',
    'sex': '-',
    'age': None,
}

datas = []
with open('gvcflist.txt') as in_f:
    for line in in_f:
        gvcf_path = line.strip()
        sample_name = os.path.basename(gvcf_path).replace('.hard-filtered.g.vcf.gz', '')
        if not utils.file_exists(gvcf_path):
            print(f'gvcf doesnt exist for {sample_name}: {gvcf_path}')
            continue
            
        if not utils.file_exists(gvcf_path + '.tbi'):
            print(f'tbi doesnt exist for {sample_name}: {gvcf_path + ".tbi"}')
            continue
        
        print(f'adding sample {sample_name} with gvcf {gvcf_path}')
        entry = default_entry.copy()
        entry.update(
            {
                's': sample_name,
                'external_id': sample_name,
                'topostproc_gvcf': gvcf_path,
                'project': 'nagim',
            }
        )
        datas.append(entry)

df = pd.DataFrame(datas)
out_fname = 'samples.tsv'
out_fpath = f'gs://cpg-nagim-test/joint_calling/{out_fname}'
df.to_csv(out_fname, sep='\t', index=False)
run_cmd(f'gsutil cp {out_fname} {out_fpath}')
print(f'Written samples TSV to {out_fpath}')
